#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Tool for importing data into database of nucleic acids from mmCIF files
#
# 2011 Ondrej Cecak <ondrej@cecak.cz>, CVUT FEL
# v1.0.1

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import subprocess
import psycopg2
import tempfile
import shutil
import re

# http://pymmlib.sourceforge.net/
from mmLib.Structure import *
from mmLib.FileIO import LoadStructure
from mmLib.FileIO import SaveStructure

# configuration
conn_string = "host='localhost' dbname='nadb' user='nadb' password='nadb'"
basedir_3dna = "/home/oc/job/fel/12-semestr/X36DIP/software/X3DNA/"
jmol_data = "/var/www/files/"


# validate command-line paramters
if len(sys.argv) != 2:
  sys.exit("USAGE: ./import.py PDBID!")

# validate database connection
try:
  link = psycopg2.connect(conn_string)
  cursor = link.cursor()
except:
  exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
  sys.exit("Database connection failed!\n ->%s" % (exceptionValue))

# make sure that PDBID is in upper case
pdbId=sys.argv[1].upper()


# download the PDB file in mmCIF format, use cache to save bandwich & time
if os.path.exists("cache/" + pdbId + ".cif"):
  print "File found in local cache, downloading is not necessary."
else:
  print "Downloading mmCIF " + pdbId + " file."
  returnCode = subprocess.call(["wget", "http://www.rcsb.org/pdb/files/" + pdbId + ".cif.gz", "-o", "/dev/null"])
  if returnCode != 0:
    sys.exit("ERROR: wget of URI http://www.rcsb.org/pdb/files/" + pdbId + ".cif.gz ended with non-zero return code.\nMake sure that you are able to fetch data from www.rcsb.org and that the inserted PDBID is correct!")

  print "Extracting."
  returnCode = subprocess.call(["gzip", "-d", pdbId + ".cif.gz"])
  if returnCode != 0:
    sys.exit("ERROR: extraction failed. The file probably wasn't downloaded correctly.\nMake sure that you are able to fetch data from www.rcsb.org and that the inserted PDBID is correct!")

  os.rename(pdbId + ".cif", "cache/" + pdbId + ".cif")
  shutil.copy("cache/" + pdbId + ".cif", jmol_data  + pdbId + ".cif")


# analyze the structure file via pymmlib
print "Parsing downloaded file."

try: 
  structure = LoadStructure(file="cache/" + pdbId + ".cif")
except:
  print "WARNING: LoadStructure failed"

# function fixNA: handle not available values correctly
def fixNA(value):
  if value == "?":
    value = None
  return value

# function fixNAreal: handle not available values correctly
def fixNAreal(value):
  if value == "---":
    value = None
  elif value == "----":
    value = None
  elif re.match("^---.*", value):
    value = None
  return value

# parse data for Biomolecule element
title = fixNA(structure.cifdb.get_table("citation")["title"])

authors = structure.cifdb.get_table("citation_author")
authorList = []
for author in authors:
  if author["citation_id"] == "primary":
    authorList.append(author["name"])

method = structure.cifdb.get_table("exptl")["method"]

dateRelease = structure.cifdb.get_table("database_PDB_rev")["date"]
dateDeposition = structure.cifdb.get_table("database_PDB_rev")["date_original"]

authorListStr = ""
for i in authorList:
  authorListStr = authorListStr + i

sequence = fixNA(structure.cifdb.get_table("entity_poly")["pdbx_seq_one_letter_code_can"])

# check, if the Biomolecule isn't already defined
cursor.execute("SELECT id FROM Biomolecule WHERE pdbCode='" + pdbId + "';")
records = cursor.fetchall()

if records:
  sys.exit("WARNING: Biomolecule with PDB ID " + pdbId + " is already defined!")

cursor.execute("INSERT INTO Biomolecule (pdbCode) VALUES (%s)", (pdbId,))
link.commit()

cursor.execute("SELECT currval('biomolecule_id_seq');")
biomoleculeId = str(cursor.fetchone()[0])

cursor.execute("INSERT INTO BiomoleculeData (biomoleculeId, name, authors, method, dateDeposition, dateRelease, sequence) VALUES (%s, %s, %s, %s, %s, %s, %s)", (biomoleculeId, title, authorListStr, method, dateDeposition, dateRelease, sequence,))
link.commit()

cursor.execute("SELECT currval('biomoleculedata_id_seq');")
biomoleculeDataId = str(cursor.fetchone()[0])

if method == "X-RAY DIFFRACTION":
  try:
    # parse data for MethodXRayData element
    resolutionLow = structure.cifdb.get_table("refine_hist")["d_res_low"]
    resolutionHigh = structure.cifdb.get_table("refine_hist")["d_res_high"]
  except:
    resolutionLow = None
    resolutionHigh = None

  try:
    method = fixNA(structure.cifdb.get_table("exptl_crystal_grow")["method"])
    temperature = fixNA(structure.cifdb.get_table("exptl_crystal_grow")["temp"])
    pH = fixNA(structure.cifdb.get_table("exptl_crystal_grow")["pH"])
  except:
    method = None
    temperature = None
    pH = None

  try:
    reflectionsAll = fixNA(structure.cifdb.get_table("refine")["ls_number_reflns_all"])
    reflectionsObs = fixNA(structure.cifdb.get_table("refine")["ls_number_reflns_obs"])
  except:
    reflectionsAll = None
    reflectionsObs = None
 
  cursor.execute("INSERT INTO MethodXRayData (biomoleculeDataId, resolutionLow, resolutionHigh, method, temperature, pH, reflectionsAll, reflectionsObs) VALUES (%s, %s, %s, %s, %s, %s, %s, %s)", (biomoleculeDataId, resolutionLow, resolutionHigh, method, temperature, pH, reflectionsAll, reflectionsObs,))
  link.commit()
elif method == "SOLUTION NMR":
  # parse data for MethodNMRData element
  try:
    fieldStrenght = fixNA(structure.cifdb.get_table("pdbx_nmr_spectrometer")["field_strength"])
  except:
    fieldStrenght = None

  try:
    numberOfModels = fixNA(structure.cifdb.get_table("pdbx_nmr_ensemble")["conformers_submitted_total_number"])
    temperature = fixNA(structure.cifdb.get_table("pdbx_nmr_exptl_sample_conditions")["temperature"])
    pressure = fixNA(structure.cifdb.get_table("pdbx_nmr_exptl_sample_conditions")["pressure"])
    pH = fixNA(structure.cifdb.get_table("pdbx_nmr_exptl_sample_conditions")["pH"])
    ionicStrength = fixNA(structure.cifdb.get_table("pdbx_nmr_exptl_sample_conditions")["ionic_strength"])
  except:
    numberOfModels = None
    temperature = None
    pressure = None
    pH = None
    ionicStrength = None

  if type(pH).__name__=='str':
    if (pH.find("-") != -1):
      phReal=pH[0:pH.find("-")]
      cursor.execute("INSERT INTO CustomField (biomoleculeId, customName, customValue) VALUES (%s, %s, %s)", (biomoleculeId, "MethodNMRData.pH.range", pH,))
      link.commit()
      pH = phReal

  cursor.execute("INSERT INTO MethodNMRData (biomoleculeDataId, fieldStrength, numberOfModels, temperature, pH, pressure, ionicStrenght) VALUES (%s, %s, %s, %s, %s, %s, %s)", (biomoleculeDataId, fieldStrenght, numberOfModels, temperature, pH, pressure, ionicStrength,))
  link.commit()
else:
  print "WARNING: unknown experiment method " + method + ", details won't be extracted!"

for chain in structure.iter_chains():
  # parse data for Chain element
  name = chain.chain_id
  sequence = str(chain[0]) + " ... " + str(chain[chain.count_fragments() - 1]);
  lenght = str(chain.count_fragments())

  cursor.execute("INSERT INTO Chain (biomoleculeId, name, sequence, lenght) VALUES (%s, %s, %s, %s)", (biomoleculeId, name, sequence,lenght,))
  link.commit()

  cursor.execute("SELECT currval('chain_id_seq');")
  chainId = str(cursor.fetchone()[0])

  for fragment in chain.iter_fragments():
    # parse data for Residue element, notice, that we have here residues + other fragments
    name = fragment.res_name
    position = fragment.fragment_id
    
    cursor.execute("INSERT INTO Residue (chainId, name, position) VALUES (%s, %s, %s)", (chainId, name, position,))
    link.commit()

    cursor.execute("SELECT currval('residue_id_seq');")
    residueId = str(cursor.fetchone()[0])

    for atom in fragment.iter_atoms():
      # parse data for Atom element
      element = atom.element
      position = atom.fragment_id
      x = str(atom.position[0])
      y = str(atom.position[1])
      z = str(atom.position[2])

      cursor.execute("INSERT INTO Atom (residueId, element, position, x, y, z) VALUES (%s, %s, %s, %s, %s, %s)", (residueId, element, position, x, y, z,))
      link.commit()

# download PDB file -- again simple cache to save bandwich & time
if os.path.exists("cache/" + pdbId + ".pdb"):
  print "File found in local cache, downloading is not necessary."
else:
  print "Downloading PDB " + pdbId + " file."
  returnCode = subprocess.call(["wget", "http://www.rcsb.org/pdb/files/" + pdbId + ".pdb.gz", "-o", "/dev/null"])
  if returnCode != 0:
    sys.exit("ERROR: wget of URI http://www.rcsb.org/pdb/files/" + pdbId + ".pdb.gz ended with non-zero return code.\nMake sure that you are able to fetch data from www.rcsb.org and that the inserted PDBID is correct!")

  print "Extracting."
  returnCode = subprocess.call(["gzip", "-d", pdbId + ".pdb.gz"])
  if returnCode != 0:
    sys.exit("ERROR: extraction failed. The file probably wasn't downloaded correctly.\nMake sure that you are able to fetch data from www.rcsb.org and that the inserted PDBID is correct!")

  os.rename(pdbId + ".pdb", "cache/" + pdbId + ".pdb")
  shutil.copy("cache/" + pdbId + ".pdb", jmol_data  + pdbId + ".pdb")

temporaryDirectory = tempfile.mkdtemp()

shutil.copy("cache/" + pdbId + ".pdb", temporaryDirectory + "/" + pdbId + ".pdb")

os.chdir(temporaryDirectory)

# perform analysis on clean, original PDB file  
os.environ["X3DNA"]=basedir_3dna
returnCode = subprocess.call([basedir_3dna + "bin/find_pair", pdbId + ".pdb", pdbId + ".inp"])
if returnCode != 0:
  print "WARNING: find_pair return non-zero status when parsing structure ID " + pdbId
returnCode = subprocess.call([basedir_3dna + "bin/analyze", pdbId + ".inp"])
if returnCode != 0:
  print "WARNING: find_pair return non-zero status when parsing structure ID " + pdbId


# get from .inp file new order of data 
conversionArray = {}
conversionArrayChain = {}
try:
  for line in open(pdbId + '.inp','r').readlines():
    data = re.split('\W+', line);
  
    # find all float numbers in the string
    parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)
  
    if (len(parameter) > 10):
      currentChain = line[line.find(">") + 1]
      conversionArrayChain[int(parameter[3])] = currentChain

      positionStart = line.find(">") + 3
      positionCurrent = positionStart
      positionEnd = line.find("_")
      numberBuilder = ""

      if (positionEnd == -1):
        print "WARNING: unidentified position of residue, skipping"
        continue

      while(positionCurrent != positionEnd):
        numberBuilder = numberBuilder + line[positionCurrent]
        positionCurrent += 1

      if(conversionArray.has_key(currentChain) == False):
	conversionArray[currentChain] = {}

      conversionArray[currentChain][int(parameter[3])] = int(re.sub(r'[^-0-9]*', "", numberBuilder.strip(".")))
except IOError:
  print "WARNING: the file " + pdbId + ".pdb wasn't parsed sucessfully -- missing " + pdbId + ".inp file."

print conversionArray

offset = 0
errors = 0
done = True
# Insert method 2: Use conversionArray to get correct position for residues
chains = {}
cursor.execute("SELECT name, id FROM chain WHERE biomoleculeId=" + biomoleculeId +";")
records = cursor.fetchall()
for record in records:
  chainName = str(record[0])
  chainId = str(record[1])
  chains[chainName] = chainId

try:
  for line in open('bp_step.par','r').readlines():
    data = re.split('\W+', line);
  
    # DATA: value from the 3DNA file
    # NAME: value picked as residue candidate for matching
  
    # manual fix for low case letters like c
    data[0] = string.upper(data[0])
  
    if (data[0] == "A") or (data[0] == "C") or (data[0] == "G") or (data[0] == "T") or (data[0] == "U"):
      # find all float numbers in the string
      parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)
     
      offset += 1
  
      currentChain=conversionArrayChain[offset]
      if (conversionArray[currentChain].has_key(offset) != True):
	print conversionArray[currentChain]
        continue
  
      try:
        cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
      except KeyError:
        # the residue is not in conversion table, very likely unknown
        continue
      records = cursor.fetchall()
  
      for record in records:
        id = str(record[0])
        name = str(record[1])
        position = str(record[2])
  
      # manual fix for residues in format like C38 instead of DC
      if (name[0] == "A") or (name[0] == "C") or (name[0] == "G") or (name[0] == "T") or (name[0] == "U"):
        name = "D" + name[0]
  
      # manual fix for residues in format like 5CM instead of DC
      if (len(name) > 1):
        if (name[1] == "A") or (name[1] == "C") or (name[1] == "G") or (name[1] == "T") or (name[1] == "U"):
          name = "D" + name[1]
  
      # manual fix for residues in format like BRU instead of DU
      if (len(name) > 2):
        if (name[2] == "A") or (name[2] == "C") or (name[2] == "G") or (name[2] == "T") or (name[2] == "U"):
          name = "D" + name[2]
  
      if "D" + data[0] == name:
        save = "yes"
      else:
        save = "no"
  
      if (save == "yes"):
        cursor.execute("INSERT INTO StructureParameter(residueId, shear, stretch, stagger, buckle, propeller, opening, shift, slide, rise, tilt, roll, twist) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", (id, parameter[0], parameter[1], parameter[2], parameter[3], parameter[4], parameter[5], parameter[6], parameter[7], parameter[8], parameter[9], parameter[10], parameter[11],))
      else:
        print "WARNING: expectation of method 2 (" + "D" + str(data[0]) + " == " + str(name) + "|" + str(id) + ", " + str(position) + "/" + str(currentChain) + ":" + str(offset) + ") about position failed, please check manualy -- possible issue at 3DNA output or incorrect mapping 3DNA:mmLIB!"
        done = False
	errors += 1
except IOError:
  print "WARNING: no base pairs found!"
  done = False

link.commit()
if ((done != False) or (errors < 3)):
  print "VERIFY: OK"
else:
  print "WARNING: all import methods failed!"

# import the rest of structural parameters
if ((done != False) or (errors < 3)):
  state=0
  inSameState=True
  stage=0
  try:
    for line in open(pdbId + '.out','r').readlines():
      # find section delimiters
      pattern = re.compile("^\*{76}$");
      if (pattern.match(line)):
        state += 1
        inSameState = False
        regexpCond=False
  
      # find known sections
      if (inSameState == False):
        if re.match("^Overlap area in Angstrom\^2.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Origin \(Ox, Oy, Oz\) and mean normal vector \(Nx, Ny, Nz\) of each base-pair.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Local base-pair helical parameters.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Classification of each dinucleotide step in a right-handed nucleic acid.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Global linear helical axis defined by equivalent C1' and RN9/YN1 atom pairs.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Main chain and chi torsion angles.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Sugar conformational parameters.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Same strand P--P and C1'--C1' virtual bond distances.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Helix radius \(radial displacement of P, O4', and C1' atoms in local.*", line):
          stage += 1
          inSameState = True
        elif re.match("^Position \(Px, Py, Pz\) and local helical axis vector \(Hx, Hy, Hz\).*", line):
          stage += 1
          inSameState = True
      
      if (inSameState == True):
        if stage==1:
          # Overlap area
          if (regexpCond == True):
            parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)

            # get current residue.id
            offset = int(parameter[0])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            cursor.execute("INSERT INTO OverlapArea(residueId, i1_i2, i1_i2Ring, i1_j2, i1_j2Ring, j1_i2, j1_i2Ring, j1_j2, j1_j2Ring) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)", (id, parameter[1], parameter[2], parameter[3], parameter[4], parameter[5], parameter[6], parameter[7], parameter[8],))
            link.commit()

          elif re.match("^.*step.*i1-i2.*i1-j2.*j1-i2.*j1-j2.*sum.*", line):
            regexpCond = True
        elif stage==2:
          # Origin and mean normal vector of each base-pair
          if (regexpCond == True):
            parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)

            # get current residue.id
            offset = int(parameter[0])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            cursor.execute("INSERT INTO OriginMeanNormalVector(residueId, Ox, Oy, Oz, Nx, Ny, Nz) VALUES (%s, %s, %s, %s, %s, %s, %s)", (id, parameter[1], parameter[2], parameter[3], parameter[4], parameter[5], parameter[6],))
            link.commit()

          elif re.match("^.*bp.*Ox.*Oy.*Oz.*Nx.*Ny.*Nz.*", line):
            regexpCond = True
        elif stage==3:
          # Local base-pair helical parameters
          if (regexpCond == True):
            parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)

            if (len(parameter) == 7):
              # get current residue.id
              offset = int(parameter[0])
              currentChain=conversionArrayChain[offset]
              if (conversionArray[currentChain].has_key(offset) != True):
                print conversionArray[currentChain]
                continue

              try:
                cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
              except KeyError:
                # the residue is not in conversion table, very likely unknown
                continue
              records = cursor.fetchall()

              for record in records:
                id = str(record[0])
                position = str(record[2])

              # insert
              cursor.execute("INSERT INTO HelicalParameter(residueId, X_disp, Y_disp, h_Rise, Incl, Tip, h_Twist) VALUES (%s, %s, %s, %s, %s, %s, %s)", (id,  parameter[1], parameter[2], parameter[3], parameter[4], parameter[5], parameter[6],))
              link.commit()

          elif re.match("^.*step.*X-disp.*Y-disp.*h-Rise.*Incl.*Tip.*h-Twist.*", line):
            regexpCond = True
        elif stage==4:
          # Classification of each dinucleotide step in a right-handed nucleic acid
          if (regexpCond == True):
            parameter = re.split(":? *", line)

            if (len(parameter) == 9):
              parameter.append(None)

            # get current residue.id
            offset = int(parameter[1])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            #cursor.execute("INSERT INTO StepClassification(residueId, Xp, Yp, Zp, XpH, YpH, ZpH) VALUES (%s, %s, %s, %s, %s, %s, %s)", (id, fixNAreal(parameter[4]), fixNAreal(parameter[5]), fixNAreal(parameter[6]), fixNAreal(parameter[7]), fixNAreal(parameter[8]), fixNAreal(parameter[9]),))
            cursor.execute("INSERT INTO StepClassification(residueId, Xp, Yp, Zp, XpH, YpH, ZpH) VALUES (%s, %s, %s, %s, %s, %s, %s)", (id, fixNAreal(parameter[3]), fixNAreal(parameter[4]), fixNAreal(parameter[5]), fixNAreal(parameter[6]), fixNAreal(parameter[7]), fixNAreal(parameter[8]),))
            link.commit()

          elif re.match("^.*step.*Xp.*Yp.*Zp.*XpH.*YpH.*ZpH.*Form.*", line):
            regexpCond = True
        elif stage==5:
          # Global linear helical axis defined by equivalent C1' and RN9/YN1 atom pairs
          if (regexpCond == True):
            parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)
            if (len(parameter) == 3):
              parameter.append(None)
              parameter.append(None)

            if (len(parameter) == 5):
              # get current residue.id
              offset = int(parameter[0])
              currentChain=conversionArrayChain[offset]
              if (conversionArray[currentChain].has_key(offset) != True):
                print conversionArray[currentChain]
                continue

              try:
                cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
              except KeyError:
                # the residue is not in conversion table, very likely unknown
                continue
              records = cursor.fetchall()

              for record in records:
                id = str(record[0])
                position = str(record[2])

              # insert
              cursor.execute("INSERT INTO C1GlobalParameter(residueId, disp, angle, twist, rise) VALUES (%s, %s, %s, %s, %s)", (id, parameter[1]  , parameter[2], parameter[3], parameter[4],))
              link.commit()

          elif re.match("^.*bp.*disp.*angle.*twist.*rise.*", line):
            regexpCond = True
        elif stage==6:
          # Main chain and chi torsion angles
          if (regexpCond == True):
            parameter = re.split(":? *", line)
            if (len(parameter) < 10):
              # don't parse rest of section after hit on first strand
              inSameState = False
            if (len(parameter) == 10):
              # get current residue.id
              offset = int(parameter[1])
              currentChain=conversionArrayChain[offset]
              if (conversionArray[currentChain].has_key(offset) != True):
                print conversionArray[currentChain]
                continue

              try:
                cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
              except KeyError:
                # the residue is not in conversion table, very likely unknown
                continue
              records = cursor.fetchall()

              for record in records:
                id = str(record[0])
                position = str(record[2])

              # insert
              cursor.execute("INSERT INTO TorsionAngle(residueId, alpha, beta, gamma, delta, epsilon, zeta, chi) VALUES (%s, %s, %s, %s, %s, %s, %s, %s)", (id, fixNAreal(parameter[3]), fixNAreal(parameter[4]), fixNAreal(parameter[5]), fixNAreal(parameter[6]), fixNAreal(parameter[7]), fixNAreal(parameter[8]), fixNAreal(parameter[9]),))
              link.commit()

          elif re.match("^.*base.*alpha.*beta.*gamma.*delta.*epsilon.*zeta.*chi.*", line):
            regexpCond = True
        elif stage==7:
          # Sugar conformational parameters
          if (regexpCond == True):
            parameter = re.findall(r"[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", line)

            if (len(parameter) < 9):
              # don't parse rest of section after hit on first strand
              inSameState = False

            if (len(parameter) == 9):
              # get current residue.id
              offset = int(parameter[0])
              currentChain=conversionArrayChain[offset]
              if (conversionArray[currentChain].has_key(offset) != True):
                print conversionArray[currentChain]
                continue

              try:
                cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
              except KeyError:
                # the residue is not in conversion table, very likely unknown
                continue
              records = cursor.fetchall()

              for record in records:
                id = str(record[0])
                position = str(record[2])

              # insert
              cursor.execute("INSERT INTO SugarConformationalParameter(residueId, tm, P) VALUES (%s, %s,  %s)", (id, parameter[6], parameter[7],))
              link.commit()

          elif re.match("^.*base.*v0.*v1.*v2.*v3.*v4.*tm.*P.*Puckering.*", line):
            regexpCond = True
        elif stage==8:
          # Same strand P--P and C1'--C1' virtual bond distances (MARK 2x)
          if (regexpCond == True):
            parameter = re.split(":? *", line)

            # get current residue.id
            offset = int(parameter[1])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            cursor.execute("INSERT INTO PCVirtualBondDistance(residueId, PP, C1C1) VALUES (%s, %s, %s)", (id, fixNAreal(parameter[3]), fixNAreal(parameter[4]),))
            link.commit()

          elif re.match("^.*step.*P--P.*C1'--C1'.*step.*P--P.*C1'--C1'.*", line):
            regexpCond = True
        elif stage==9:
          if (regexpCond == True):
            parameter = re.split(":? *", line)

            # get current residue.id
            offset = int(parameter[1])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            cursor.execute("INSERT INTO HelixRadiusRadialDisplacement(residueId, P, O4, C1) VALUES (%s, %s, %s, %s)", (id, fixNAreal(parameter[3]), fixNAreal(parameter[4]), fixNAreal(parameter[5]),))
            link.commit()

          elif re.match("^.*step.*P.*O4'.*C1'.*P.*O4'.*C1'", line):
            regexpCond = True
        elif stage==10:
          # Position and local helical axis vector 
          if (regexpCond == True):
            parameter = re.split(":? *", line)

            # get current residue.id
            offset = int(parameter[1])
            currentChain=conversionArrayChain[offset]
            if (conversionArray[currentChain].has_key(offset) != True):
              print conversionArray[currentChain]
              continue

            try:
              cursor.execute("SELECT id, name, position FROM residue WHERE chainId=" + chains[currentChain]  + " AND position=" + str(conversionArray[currentChain][offset]) +";")
            except KeyError:
              # the residue is not in conversion table, very likely unknown
              continue
            records = cursor.fetchall()

            for record in records:
              id = str(record[0])
              position = str(record[2])

            # insert
            cursor.execute("INSERT INTO PositionAndLocalHelicalAxisVector(residueId, Px, Py, Pz, Hx, Hy, Hz) VALUES (%s, %s, %s, %s, %s, %s, %s)", (id, fixNAreal(parameter[3]), fixNAreal(parameter[4]), fixNAreal(parameter[5]), fixNAreal(parameter[6]), fixNAreal(parameter[7]), fixNAreal(parameter[8]),))
            link.commit()

          elif re.match("^.*step.*Px.*Py.*Pz.*Hx.*Hy.*Hz.*", line):
            regexpCond = True
  
      #if ((stage == 10) and (inSameState == True)):
      #  print "DEBUG: (" + str(stage) + "/" + str(state) + "): " + line
  
  except IOError:
    print "WARNING: no extended parameters found!"
    done = False

shutil.rmtree(temporaryDirectory)

link.close()

